//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/Scale.cpp
//! @brief     Implements interface Scale.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Axis/Scale.h"
#include "Base/Util/Assert.h"
#include "Base/Util/StringUtil.h"
#include <iomanip>
#include <iostream>
#include <numbers>
#include <stdexcept>
using std::numbers::pi;

namespace {

trafo_map_t::const_iterator find(std::string label, const trafo_map_t& trafos)
{
    for (trafo_map_t::const_iterator e = trafos.cbegin(); e != trafos.cend(); ++e)
        if (e->first == label)
            return e;
    return trafos.cend();
}

} // namespace


Scale::Scale(std::vector<Bin1D> bins, trafo_map_t trafos)
    : m_bins(std::move(bins))
    , m_trafos(std::move(trafos))
{
    if (size() == 0)
        throw std::runtime_error("Scale constructor called with no bins");
    for (size_t i = 0; i < size() - 1; ++i) {
        if (bin(i).upperBound() > bin(i + 1).lowerBound())
            throw std::runtime_error("Scale constructor called with overlapping bins");
        if (bin(i) == bin(i + 1))
            throw std::runtime_error("Scale constructor called with repeating bin(s)");
    }
    if (isScan()) {
        for (const Bin1D& b : m_bins)
            if (b.binSize() != 0)
                throw std::runtime_error("Finite bin(s) in scan");
    } else {
        for (const Bin1D& b : m_bins)
            if (b.binSize() == 0)
                throw std::runtime_error("Empty bin(s) in sweep");
    }
}

Scale::Scale(const Coordinate& coord, std::vector<Bin1D> bins)
    : Scale(bins, {{coord.label(), [](double x) { return x; }}})
{
}

Scale* Scale::clone() const
{
    return new Scale(m_bins, m_trafos);
}

void Scale::setTrafos(trafo_map_t trafos)
{
    m_trafos = std::move(trafos);
}

std::string Scale::availableLabels() const
{
    std::vector<std::string> labels;
    for (const auto& entry : m_trafos)
        labels.push_back(entry.first);
    return Base::String::join(labels, ", ");
}

std::string Scale::axisLabel() const
{
    return m_trafos.cbegin()->first;
}

std::string Scale::unit() const
{
    return Coordinate(axisLabel()).unit();
}

size_t Scale::size() const
{
    return m_bins.size();
}

double Scale::min() const
{
    return m_bins.front().lowerBound();
}

double Scale::max() const
{
    return m_bins.back().upperBound();
}

std::pair<double, double> Scale::bounds() const
{
    return {min(), max()};
}

bool Scale::rangeComprises(double value) const
{
    return value >= min() && value < max();
}

double Scale::span() const
{
    return max() - min();
}

double Scale::center() const
{
    return (max() + min()) / 2;
}

const Bin1D& Scale::bin(size_t i) const
{
    return m_bins.at(i);
}

double Scale::binCenter(size_t i) const
{
    return bin(i).center();
}

std::vector<double> Scale::binCenters() const
{
    std::vector<double> result;
    for (const Bin1D& b : m_bins)
        result.emplace_back(b.center());
    return result;
}

size_t Scale::closestIndex(double value) const
{
    for (size_t i = 0; i < size() - 1; ++i)
        if (value < (bin(i).upperBound() + bin(i + 1).lowerBound()) / 2)
            return i;
    return size() - 1;
}

bool Scale::isEquiDivision() const
{
    const size_t N = size();
    for (size_t i = 0; i < N; ++i) {
        const Bin1D& b = bin(i);
        // exactly replicate the computation of bin bounds in the EquiDivision factory function
        if (b.lowerBound() != (N - i) * (min() / N) + i * (max() / N)
            || b.upperBound() != (N - i - 1) * (min() / N) + (i + 1) * (max() / N))
            return false;
    }
    return true;
}

bool Scale::isEquiScan() const
{
    const size_t N = size();
    ASSERT(N);
    if (N == 1)
        return !bin(0).binSize();
    for (size_t i = 0; i < N; ++i) {
        const Bin1D& b = bin(i);
        if (b.binSize())
            return false;
        // exactly replicate the computation of bin bounds in the EquiDivision factory function
        if (b.lowerBound() != (N - 1 - i) * (min() / (N - 1)) + i * (max() / (N - 1)))
            return false;
    }
    return true;
}

bool Scale::isScan() const
{
    for (const Bin1D& b : bins())
        if (b.binSize())
            return false;
    return true;
}

Scale Scale::clipped(double lower, double upper) const
{
    if (lower > upper)
        throw std::runtime_error("Scale::clipped called with invalid bounds (lower > upper)");
    std::vector<Bin1D> out_bins;
    const bool is_scan = isScan();
    for (const Bin1D& b : m_bins)
        if (auto bc = b.clipped_or_nil(lower, upper))
            if (is_scan || bc.value().binSize() > 0)
                out_bins.emplace_back(bc.value());
    return {out_bins, m_trafos};
}

Scale Scale::clipped(std::pair<double, double> bounds) const
{
    return clipped(bounds.first, bounds.second);
}

bool Scale::operator==(const Scale& other) const
{
    return axisLabel() == other.axisLabel() && m_bins == other.m_bins;
}

std::ostream& operator<<(std::ostream& ostr, const Scale& ax)
{
    size_t N = ax.size();
    ASSERT(N > 0);

    ostr << std::setprecision(15);

    if (ax.isScan()) {
        ostr << "ListScan(\"" << ax.axisLabel() << "\", [";
        for (double v : ax.binCenters())
            ostr << v << ",";
        ostr << "])";
        return ostr;
    }

    if (ax.isEquiDivision()) {
        ostr << "EquiDivision(\"" << ax.axisLabel() << "\", " << ax.size() << ", " << ax.min()
             << ", " << ax.max() << ")";
        return ostr;
    }

    ostr << "GenericScale(\"" << ax.axisLabel() << "\", [";
    for (const Bin1D& b : ax.bins())
        ostr << b.lowerBound() << "," << b.upperBound() << ",";
    ostr << "])";
    return ostr;
}

Scale Scale::plottableScale(std::string label) const
{
    ASSERT(!m_trafos.empty());
    trafo_map_t::const_iterator entry;
    if (label.empty()) {
        entry = m_trafos.cbegin();
        label = entry->first;
    } else {
        entry = find(label, m_trafos);
        if (entry == m_trafos.cend())
            throw std::runtime_error("Scale::plottableScale called with unknown label '" + label
                                     + "';" + " available labels are: " + availableLabels());
    }
    const trafo_t& trafo = entry->second;
    std::vector<Bin1D> outvector;
    for (const Bin1D b : m_bins) {
        double bmi = trafo(b.lowerBound());
        double bma = trafo(b.upperBound());
        outvector.emplace_back(Bin1D::FromTo(bmi, bma));
    }
    return {label, outvector};
}

Scale Scale::plottableScale(const Coordinate& coord, const trafo_t& axTrafo) const
{
    std::vector<Bin1D> outvector;
    for (const Bin1D b : m_bins) {
        double bmi = axTrafo(b.lowerBound());
        double bma = axTrafo(b.upperBound());
        outvector.emplace_back(Bin1D::FromTo(bmi, bma));
    }
    return {coord.label(), outvector};
}
